﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Workflow.Runtime;

namespace Hardcodet.WF.Util.Hosting
{
  /// <summary>
  /// Manages a set of workflows.
  /// </summary>
  public class WorkflowRuntimeManager : IDisposable
  {
    #region properties

    /// <summary>
    /// The managed workflow runtime.
    /// </summary>
    public WorkflowRuntime Runtime { get; private set; }

    /// <summary>
    /// Gets the managed workflows, stored by their individual workflow instance
    /// IDs.
    /// </summary>
    public WorkflowWrapperCollection Workflows { get; private set; }

    #endregion

    #region events

    /// <summary>
    /// Publishes status changed of the internal <see cref="Runtime"/>.
    /// </summary>
    public event EventHandler<WorkflowRuntimeEventArgs> RuntimeStatusChanged;


    /// <summary>
    /// Publishes status changed of a workflow that is managed by the internal
    /// <see cref="Runtime"/>.
    /// </summary>
    public event EventHandler<WorkflowLogEventArgs> WorkflowStatusChanged;


    /// <summary>
    /// Raises the <see cref="WorkflowStatusChanged"/> event for a given workflow.
    /// </summary>
    /// <param name="eventArgs"></param>
    /// <param name="state"></param>
    private void RaiseWorkflowStatusEvent(WorkflowEventArgs eventArgs, WorkflowState state)
    {
      if (WorkflowStatusChanged != null)
      {
        WorkflowInstanceWrapper wrapper = TryGetWorkflow(eventArgs.WorkflowInstance.InstanceId);

        //if the workflow that raised the event is not managed internally, ignore it
        if (wrapper == null) return;

        string msg = "Workflow {0} changed state to: {1}.";
        msg = String.Format(msg, wrapper.WorkflowInstance.InstanceId, state);

        WorkflowLogEventArgs args = new WorkflowLogEventArgs(wrapper, state, msg);
        WorkflowStatusChanged(this, args);
      }
    }

    #endregion

    #region construction

    /// <summary>
    /// Inits the class with the workflow runtime to be managed.
    /// </summary>
    /// <param name="runtime">The managed workflow runtime.</param>
    /// <exception cref="ArgumentNullException">If <paramref name="runtime"/>
    /// is a null reference.</exception>
    public WorkflowRuntimeManager(WorkflowRuntime runtime)
    {
      if (runtime == null) throw new ArgumentNullException("runtime");
      Runtime = runtime;

      //init auto properties
      Workflows = new WorkflowWrapperCollection();

      //subscribe to events
      RegisterRuntimeEventListeners(runtime);
    }

    #endregion

    #region runtime event listeners

    /// <summary>
    /// Registers workflow and runtime event listeners.
    /// </summary>
    /// <param name="runtime">The managed workflow runtime.</param>
    private void RegisterRuntimeEventListeners(WorkflowRuntime runtime)
    {
      //runtime events
      runtime.Started += OnRuntimeStateChanged;
      runtime.Stopped += OnRuntimeStateChanged;

      //workflow events
      runtime.WorkflowCreated += OnWorkflowCreated;
      runtime.WorkflowStarted += OnWorkflowStarted;
      runtime.WorkflowLoaded += OnWorkflowLoaded;
      runtime.WorkflowIdled += OnWorkflowIdled;
      runtime.WorkflowPersisted += OnWorkflowPersisted;
      runtime.WorkflowResumed += OnWorkflowResumed;
      runtime.WorkflowUnloaded += OnWorkflowUnloaded;
      runtime.WorkflowCompleted += OnWorkflowCompleted;
      runtime.WorkflowTerminated += OnWorkflowTerminated;
      runtime.WorkflowAborted += OnWorkflowAborted;
      runtime.WorkflowSuspended += OnWorkflowSuspended;
    }


    /// <summary>
    /// Raises the <see cref="RuntimeStatusChanged"/> event in order to propagate
    /// a status change of the runtime itself.
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    private void OnRuntimeStateChanged(object sender, System.Workflow.Runtime.WorkflowRuntimeEventArgs e)
    {
      if (RuntimeStatusChanged != null)
      {
        RuntimeStatusChanged(this, new WorkflowRuntimeEventArgs(Runtime, e.IsStarted));
      }
    }


    private void OnWorkflowCreated(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Created);
    }


    private void OnWorkflowLoaded(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Loaded);
    }


    private void OnWorkflowIdled(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Idled);
    }


    private void OnWorkflowPersisted(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Persisted);
    }


    private void OnWorkflowResumed(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Resumed);
    }


    private void OnWorkflowUnloaded(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Unloaded);
    }


    private void OnWorkflowStarted(object sender, WorkflowEventArgs e)
    {
      RaiseWorkflowStatusEvent(e, WorkflowState.Started);
    }


    private void OnWorkflowTerminated(object sender, WorkflowTerminatedEventArgs e)
    {
      WorkflowInstanceWrapper wrapper = TryGetWorkflow(e.WorkflowInstance.InstanceId);
      if (wrapper != null)
      {
        wrapper.Exception = e.Exception;
        wrapper.StopWaiting();
      }

      RaiseWorkflowStatusEvent(e, WorkflowState.Terminated);
    }


    private void OnWorkflowCompleted(object sender, WorkflowCompletedEventArgs e)
    {
      WorkflowInstanceWrapper wrapper = TryGetWorkflow(e.WorkflowInstance.InstanceId);
      if (wrapper != null)
      {
        wrapper.OutputParameters = e.OutputParameters;
        wrapper.StopWaiting();
      }

      RaiseWorkflowStatusEvent(e, WorkflowState.Completed);
    }


    private void OnWorkflowAborted(object sender, WorkflowEventArgs e)
    {
      WorkflowInstanceWrapper wrapper = TryGetWorkflow(e.WorkflowInstance.InstanceId);
      if (wrapper != null)
      {
        wrapper.StopWaiting();
      }

      RaiseWorkflowStatusEvent(e, WorkflowState.Aborted);
    }


    private void OnWorkflowSuspended(object sender, WorkflowSuspendedEventArgs e)
    {
      WorkflowInstanceWrapper wrapper = TryGetWorkflow(e.WorkflowInstance.InstanceId);
      if (wrapper != null)
      {
        wrapper.ReasonSuspended = e.Error;
      }

      RaiseWorkflowStatusEvent(e, WorkflowState.Suspended);
    }

    #endregion

    #region add / remove / get workflows

    /// <summary>
    /// Creates a workflow instance, encapsulates it in a <see cref="WorkflowInstanceWrapper"/>
    /// class and caches the wrapper in the <see cref="Workflows"/> collection.
    /// </summary>
    /// <param name="workflowType">The type of the workflow to be created.</param>
    /// <param name="workflowParameters">Workflow parameters, if any. May be null.</param>
    /// <returns>The created wrapper class.</returns>
    public WorkflowInstanceWrapper CreateWorkflow(Type workflowType, Dictionary<string, object> workflowParameters)
    {
      return CreateWorkflow(workflowType, workflowParameters, false);
    }


    /// <summary>
    /// Creates a workflow instance, encapsulates it in a <see cref="WorkflowInstanceWrapper"/>
    /// class and caches the wrapper in the <see cref="Workflows"/> collection.
    /// </summary>
    /// <param name="workflowType">The type of the workflow to be created.</param>
    /// <param name="workflowParameters">Workflow parameters, if any. May be null.</param>
    /// <param name="autoStart">Whether to automatically start the created workflow instance, once
    /// it has been locally cached.</param>
    /// <returns>The created wrapper class.</returns>
    public WorkflowInstanceWrapper CreateWorkflow(Type workflowType, Dictionary<string, object> workflowParameters,
                                                  bool autoStart)
    {
      //create and wrap workflow
      WorkflowInstance workflow = Runtime.CreateWorkflow(workflowType, workflowParameters);
      WorkflowInstanceWrapper wrapper = ManageWorkflow(workflow);

      //automatically start the workflow, if requested
      if (autoStart) workflow.Start();
      return wrapper;
    }


    /// <summary>
    /// Assigns an existing workflow that belongs to the internal <see cref="Runtime"/>
    /// to the manager's <see cref="Workflows"/> collection, encapsulated in a
    /// <see cref="WorkflowInstanceWrapper"/> instance.
    /// </summary>
    /// <returns>The created wrapper class.</returns>
    /// <exception cref="ArgumentNullException">If <paramref name="workflow"/>
    /// is a null reference.</exception>
    /// <exception cref="InvalidOperationException">If the submitted workflow's runtime
    /// is different than the internal <see cref="Runtime"/>.</exception>
    /// <exception cref="ArgumentException">If the workflow (workflow with the same
    /// <see cref="WorkflowInstance.InstanceId"/>) is already being managed.</exception>
    public WorkflowInstanceWrapper ManageWorkflow(WorkflowInstance workflow)
    {
      if (workflow == null) throw new ArgumentNullException("workflow");

      if (!ReferenceEquals(workflow.WorkflowRuntime, Runtime))
      {
        string msg = "The submitted workflow's runtime ({0}) does not match the runtime of the manger class ({1}.";
        msg = String.Format(msg, workflow.WorkflowRuntime.Name, Runtime.Name);
        throw new InvalidOperationException(msg);
      }

      if (Workflows.Contains(workflow.InstanceId))
      {
        string msg = "Cannot add workflow {0} - a workflow with the same instance ID is already managed.";
        msg = String.Format(msg, workflow.InstanceId);
        throw new ArgumentException(msg);
      }

      //wrap workflow
      WorkflowInstanceWrapper wrapper = new WorkflowInstanceWrapper(workflow);

      //store the wrapper in the internal workflows collection
      Workflows.Add(wrapper);
      return wrapper;
    }


    /// <summary>
    /// Removes a given workflow from the <see cref="Workflows"/>
    /// collection.
    /// </summary>
    /// <param name="wrapper">The wrapper class that encapsulates
    /// the workflow to be removed.</param>
    /// <returns>True if the element is successfully removed; otherwise, false.
    /// This method also returns false if instance ID of the workflow is not found in the
    /// <see cref="Workflows"/> collection.</returns>
    public bool RemoveWorkflow(WorkflowInstanceWrapper wrapper)
    {
      return Workflows.Remove(wrapper.WorkflowInstance.InstanceId);
    }


    /// <summary>
    /// Removes a given workflow from the <see cref="Workflows"/>
    /// collection by its <see cref="WorkflowInstance.InstanceId"/>.
    /// </summary>
    /// <param name="workflowId">The identifier of the workflow
    /// to be removed.</param>
    /// <returns>True if the element is successfully removed; otherwise, false.
    /// This method also returns false if instance ID of the workflow is not found in the
    /// <see cref="Workflows"/> collection.</returns>
    public bool RemoveWorkflow(Guid workflowId)
    {
      return Workflows.Remove(workflowId);
    }


    /// <summary>
    /// Tries to retrieve a workflow from the <see cref="Workflows"/>
    /// collection through the <see cref="WorkflowInstance.InstanceId"/>
    /// of a managed workflow.
    /// </summary>
    /// <param name="workflowId">A guid that identifies a manged
    /// <see cref="WorkflowInstance"/>.</param>
    /// <returns>The cached workflow instance wrapper, if available.
    /// Otherwise null.</returns>
    public WorkflowInstanceWrapper TryGetWorkflow(Guid workflowId)
    {
      if (!Workflows.Contains(workflowId)) return null;
      return Workflows[workflowId];
    }

    #endregion

    /// <summary>
    /// Waits for all the managed workflows to receive a signal,
    /// using an given timeout in milliseconds.
    /// </summary>
    /// <param name="millisecondsTimeout"></param>
    /// <returns>True when every workflow has received a signal within the
    /// specified period; otherwise false.</returns>
    public bool WaitAll(int millisecondsTimeout)
    {
      if (Workflows.Count == 0) return true;

      //get all handles from the managed workflows
      ManualResetEvent[] handles = (from wrapper in Workflows select wrapper.WaitHandle).ToArray();
      return WaitHandle.WaitAll(handles, millisecondsTimeout);
    }


    /// <summary>
    /// Stops and disposes the underlying <see cref="Runtime"/> and clears
    /// the <see cref="Workflows"/> collection.
    /// </summary>
    public void Dispose()
    {
      Runtime.StopRuntime();
      Runtime.Dispose();

      Workflows.Clear();
    }
  }
}